/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger;

import org.intellij.plugins.xsltDebugger.rt.engine.Debugger;
import org.intellij.plugins.xsltDebugger.rt.engine.remote.RemoteDebuggerClient;
import org.intellij.plugins.xsltDebugger.ui.DebuggerTabComponent;

import com.intellij.debugger.DebuggerManagerEx;
import com.intellij.debugger.engine.DebugProcess;
import com.intellij.debugger.impl.DebuggerSession;
import com.intellij.execution.process.ProcessHandler;
import com.intellij.execution.process.ProcessOutputTypes;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ModalityState;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.rmi.ConnectException;
import java.rmi.NotBoundException;

/**
 * Establishes the debugger-connection to the started XSLT-process, starts the debugger-session
 * and attaches the debugger-UI.
 */
class DebuggerConnector implements Runnable {

    private final Project myProject;
    private final ProcessHandler myProcess;
    private final int myPort;

    public DebuggerConnector(Project project, ProcessHandler process, int port) {
        myProject = project;
        myProcess = process;
        myPort = port;
    }

    public void run() {
        final Debugger client = connect();
        if (client == null) {
            // client will be null if the process terminated prematurely for some reason. no need for an error message
            if (!myProcess.isProcessTerminated()) {
                myProcess.notifyTextAvailable("Failed to connect to debugged process. Terminating.\n", ProcessOutputTypes.SYSTEM);
                myProcess.destroyProcess();
            }
            return;
        }

        XsltBreakpointManager.getInstance(myProject).initBreakpoints(myProcess, client);

        final XsltDebuggerSession session = XsltDebuggerSession.create(myProject, myProcess, client);
        DebuggerTabComponent.attachDebugger(myProcess, session);
        session.start();
    }

    @Nullable
    private Debugger connect() {
        final DebuggerSession[] session = new DebuggerSession[1];
        ApplicationManager.getApplication().invokeAndWait(new Runnable() {
            public void run() {
                final DebuggerManagerEx debuggerManager = DebuggerManagerEx.getInstanceEx(myProject);
                final DebugProcess process = debuggerManager.getDebugProcess(myProcess);
                session[0] = debuggerManager.getSession(process);
            }
        }, ModalityState.NON_MODAL);

        if (session[0] == null) return null;

        for (int i = 0; i < 10; i++) {
            try {
                final RemoteDebuggerClient realClient = new RemoteDebuggerClient(myPort);
                myProcess.notifyTextAvailable("Connected to XSLT debugger on port " + myPort + "\n", ProcessOutputTypes.SYSTEM);

                return (Debugger)EDTGuard.create(realClient, myProcess, session[0]);
            } catch (ConnectException e) {
                try {
                    Thread.sleep(500);
                } catch (InterruptedException e1) {
                    break;
                }
            } catch (NotBoundException e) {
                try {
                    Thread.sleep(200);
                } catch (InterruptedException e1) {
                    break;
                }
            } catch (IOException e) {
                Logger.getInstance(getClass().getName()).info("Could not connect to debugger", e);
                break;
            }
        }
        return null;
    }
}
