/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.plugins.xsltDebugger;

import org.intellij.plugins.xsltDebugger.rt.engine.Breakpoint;
import org.intellij.plugins.xsltDebugger.rt.engine.BreakpointManager;
import org.intellij.plugins.xsltDebugger.rt.engine.local.BreakpointManagerImpl;

import com.intellij.debugger.DebuggerManagerEx;
import com.intellij.debugger.SourcePosition;
import com.intellij.debugger.engine.DebugProcess;
import com.intellij.debugger.engine.DebugProcessImpl;
import com.intellij.debugger.impl.DebuggerSession;
import com.intellij.debugger.ui.breakpoints.BreakpointCategory;
import com.intellij.debugger.ui.breakpoints.LineBreakpoint;
import com.intellij.lang.ASTNode;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.markup.RangeHighlighter;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Key;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.util.PsiTreeUtil;
import com.intellij.psi.xml.XmlChildRole;
import com.intellij.psi.xml.XmlElement;
import com.intellij.psi.xml.XmlTag;
import com.intellij.psi.xml.XmlToken;
import com.intellij.psi.xml.XmlTokenType;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import com.sun.jdi.ReferenceType;

import javax.swing.*;

/**
 * Created by IntelliJ IDEA.
 * User: sweinreuter
 * Date: 26.05.2007
 */
public class XsltBreakpoint extends LineBreakpoint {
    public static final Key<BreakpointManager> MANAGER = Key.create("BREAKPOINT_MANAGER");

    @NonNls
    public static final Key<LineBreakpoint> CATEGORY = BreakpointCategory.lookup("xslt_breakpoints");

    public boolean HAS_CONDITION = false;
    public String CONDITION_EXPR = "";
    public boolean HAS_LOG_MSG = false;
    public String LOG_EXPR = "";
    public boolean HAS_TRACE_MSG = false;
    public String TRACE_EXPR = "";
    public boolean SUSPEND = false;

    XsltBreakpoint(Project project) {
        super(project);
    }

    private XsltBreakpoint(Project project, Document document, int line) {
        super(project, createHighlighter(project, document, line));
    }

    protected Icon getSetIcon() {
        return XsltBreakpointFactory.getIconStatic();
    }

    protected Icon getDisabledIcon() {
        return XsltBreakpointFactory.getDisabledIconStatic();
    }

    @Override
    protected void createRequestForPreparedClass(DebugProcessImpl debugProcess, ReferenceType classType) {
    }

    @Override
    public void createRequest(@NotNull DebugProcessImpl debugProcess) {
        synchronized (MANAGER) {
            BreakpointManager manager = debugProcess.getUserData(MANAGER);
            if (manager == null) {
                manager = new BreakpointManagerImpl();
                debugProcess.putUserData(MANAGER, manager);
            }
            final SourcePosition position = getSourcePosition();
            if (position != null) {
                setBreakpoint(manager, position);
            }
        }
        updateUI();
    }

    private static String position2Uri(@NotNull SourcePosition position) {
        return VfsUtil.virtualToIoFile(position.getFile().getVirtualFile()).toURI().toASCIIString();
    }

    private synchronized void setBreakpoint(@NotNull BreakpointManager manager, @NotNull SourcePosition position) {
        try {
            final Breakpoint breakpoint = manager.setBreakpoint(position2Uri(position), getActualLineNumber(position));
            breakpoint.setEnabled(ENABLED);
            breakpoint.setCondition(HAS_CONDITION ? CONDITION_EXPR : null);
            breakpoint.setLogMessage(HAS_LOG_MSG ? LOG_EXPR : null);
            breakpoint.setTraceMessage(HAS_TRACE_MSG ? TRACE_EXPR : null);
            breakpoint.setSuspend(SUSPEND);
        } catch (Exception e1) {
            // hmm..
        }
    }

    static int getActualLineNumber(SourcePosition position) {
        final PsiElement element = findContextElement(position);
        if (element != null) {
            // TODO: check if this needs to be relaxed
            if (element instanceof XmlToken && ((XmlToken)element).getTokenType() == XmlTokenType.XML_START_TAG_START) {
                final Document document = PsiDocumentManager.getInstance(element.getProject()).getDocument(position.getFile());
                if (document != null) {
                    if (document.getLineNumber(element.getTextRange().getStartOffset()) == position.getLine()) {
                        final XmlTag tag = PsiTreeUtil.getParentOfType(element, XmlTag.class, false);
                        if (tag != null) {
                            final ASTNode node = tag.getNode();
                            assert node != null;
                            // TODO: re-check if/when Xalan is supported
                            final ASTNode end = XmlChildRole.START_TAG_END_FINDER.findChild(node);
                            if (end != null) {
                                return document.getLineNumber(end.getTextRange().getEndOffset()) + 1;
                            } else {
                                final ASTNode end2 = XmlChildRole.EMPTY_TAG_END_FINDER.findChild(node);
                                if (end2 != null) {
                                    return document.getLineNumber(end2.getTextRange().getEndOffset()) + 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        return -1;
    }

    public boolean isValid() {
        return getActualLineNumber(getSourcePosition()) != -1;
    }

    @Override
    public void delete() {
        super.delete();

        synchronized (MANAGER) {
            for (DebuggerSession debuggerSession : DebuggerManagerEx.getInstanceEx(myProject).getSessions()) {
                final DebugProcessImpl process = debuggerSession.getProcess();
                if (process == null || !process.isAttached()) {
                    continue;
                }
                try {
                    deleteRequest(process);
                } catch (Exception e1) {
                    Logger.getInstance(getClass().getName()).debug("Failed to clear breakpoint for " + process, e1);
                }
            }
        }
    }

    private void deleteRequest(@NotNull DebugProcessImpl process) {
        final BreakpointManager manager = process.getUserData(MANAGER);
        if (manager != null) {
            final SourcePosition position = getSourcePosition();
            manager.removeBreakpoint(position2Uri(position), getActualLineNumber(position));
        }
    }

    @Override
    public void processClassPrepare(DebugProcess debugProcess, ReferenceType classType) {
    }

    @NotNull
    @Override
    public String getDescription() {
        return getDisplayName();
    }

    @NotNull
    public String getDisplayName() {
        final RangeHighlighter highlighter = getHighlighter();
        final int lineNumber = (highlighter.getDocument().getLineNumber(highlighter.getStartOffset()) + 1);
        return "Line " + lineNumber + " in " + getSourcePosition().getFile().getName();
    }

    @Nullable
    public static XsltBreakpoint createEx(Project project, Document document, int line) {
        return (XsltBreakpoint)new XsltBreakpoint(project, document, line).init();
    }

    @NotNull
    @Override
    public Key<LineBreakpoint> getCategory() {
        return CATEGORY;
    }

    @Nullable
    public PsiElement findContextElement() {
        return findContextElement(getSourcePosition());
    }

    @Nullable
    private static PsiElement findContextElement(SourcePosition position) {
        if (position == null) {
            return null;
        }
        PsiElement contextElement = position.getFile().findElementAt(position.getOffset());
        while (contextElement != null && !(contextElement instanceof XmlElement)) {
            contextElement = PsiTreeUtil.nextLeaf(contextElement);
        }
        return contextElement;
    }
}
