/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.intellij.plugins.xsltDebugger;

import org.intellij.plugins.xsltDebugger.rt.engine.Breakpoint;
import org.intellij.plugins.xsltDebugger.rt.engine.BreakpointManager;
import org.intellij.plugins.xsltDebugger.rt.engine.Debugger;

import com.intellij.debugger.DebuggerInvocationUtil;
import com.intellij.debugger.DebuggerManagerEx;
import com.intellij.debugger.engine.DebugProcess;
import com.intellij.debugger.engine.requests.RequestManagerImpl;
import com.intellij.execution.process.ProcessHandler;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.EditorFactory;
import com.intellij.openapi.editor.event.EditorEventMulticaster;
import com.intellij.openapi.editor.event.EditorMouseAdapter;
import com.intellij.openapi.editor.event.EditorMouseEvent;
import com.intellij.openapi.editor.event.EditorMouseEventArea;
import com.intellij.openapi.editor.event.EditorMouseListener;
import com.intellij.openapi.editor.markup.MarkupEditorFilterFactory;
import com.intellij.openapi.fileEditor.FileEditor;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.fileEditor.TextEditor;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.openapi.fileTypes.StdFileTypes;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiFile;
import org.jetbrains.annotations.NotNull;

import org.intellij.lang.xpath.xslt.XsltSupport;

import java.awt.event.MouseEvent;
import java.util.List;

/**
 * Manages XSLT breakpoints:
 * - Setting breakpoints via mouse-click on the editor gutter
 * - programmatic toggle of breakpoints (used by hooked keyboard action to set breakpoints)
 * - transfers settings between IDEA-breakpoints and those in the debugged process 
 */
public class XsltBreakpointManager implements ProjectComponent {
    private final Project myProject;
    private final DebuggerManagerEx myDebuggerManager;

    public XsltBreakpointManager(Project project, DebuggerManagerEx debuggerManager) {
        myProject = project;
        myDebuggerManager = debuggerManager;
    }

    public void initComponent() {
    }

    public void disposeComponent() {
    }

    @NotNull
    public String getComponentName() {
        return "XsltBreakpointManager";
    }

    public void projectOpened() {
        init();
    }

    public void projectClosed() {
    }

    public boolean toggleBreakpoint(Editor editor) {
        final int offset = editor.getCaretModel().getOffset();
        return toggleBreakpoint(editor, editor.getDocument().getLineNumber(offset));
    }

    public boolean toggleBreakpoint(Editor editor, final int line) {
        if (!canSetXsltBreakpoint(editor)) return false;

        final Document document = editor.getDocument();
        int offset = editor.getCaretModel().getOffset();
        final int editorLine = document.getLineNumber(offset);
        if (editorLine != line) {
            if (line < 0 || line >= document.getLineCount()) {
                return false;
            }
            offset = document.getLineStartOffset(line);
        }

        final com.intellij.debugger.ui.breakpoints.BreakpointManager bm =
                myDebuggerManager.getBreakpointManager();

        com.intellij.debugger.ui.breakpoints.Breakpoint breakpoint =
                bm.findBreakpoint(document, offset, XsltBreakpoint.CATEGORY);

        if (breakpoint == null) {
            breakpoint = XsltBreakpoint.createEx(myProject, document, line);
            if (breakpoint != null) {
                bm.addBreakpoint(breakpoint);
                RequestManagerImpl.createRequests(breakpoint);
            }
        } else {
            bm.removeBreakpoint(breakpoint);
        }

        return true;
    }

    public boolean canSetXsltBreakpoint(Editor editor) {
        final PsiFile psiFile = PsiDocumentManager.getInstance(myProject).getPsiFile(editor.getDocument());
        if (psiFile == null) {
            return false;
        }
        final FileType fileType = psiFile.getFileType();
        if (fileType != StdFileTypes.XML) {
            return false;
        } else if (!XsltSupport.isXsltFile(psiFile)) {
            return false;
        }
        return true;
    }

    private void init() {
        // this is a really ugly hack, but there's no (suitable) API that allows to set custom breakpoints.
        final EditorMouseListener editorMouseListener = new MyMouseAdapter();
        final EditorEventMulticaster eventMulticaster = EditorFactory.getInstance().getEventMulticaster();
        eventMulticaster.addEditorMouseListener(editorMouseListener, myProject);
    }

    public void initBreakpoints(ProcessHandler process1, @NotNull Debugger client) {
        synchronized (XsltBreakpoint.MANAGER) {
            final DebugProcess process = myDebuggerManager.getDebugProcess(process1);
            if (process == null) return;

            final BreakpointManager mgr = process.getUserData(XsltBreakpoint.MANAGER);
            final BreakpointManager manager = client.getBreakpointManager();
            if (mgr == null) {
                process.putUserData(XsltBreakpoint.MANAGER, manager);
            } else if (mgr != manager) {
                final List<Breakpoint> breakpoints = mgr.getBreakpoints();
                for (Breakpoint breakpoint : breakpoints) {
                    final Breakpoint bp = manager.setBreakpoint(breakpoint.getUri(), breakpoint.getLine());
                    bp.setEnabled(breakpoint.isEnabled());
                    bp.setLogMessage(breakpoint.getLogMessage());
                    bp.setTraceMessage(breakpoint.getTraceMessage());
                    bp.setCondition(breakpoint.getCondition());
                    bp.setSuspend(breakpoint.isSuspend());
                }
                process.putUserData(XsltBreakpoint.MANAGER, manager);
            }
        }
    }

    public void detach(ProcessHandler p) {
        synchronized (XsltBreakpoint.MANAGER) {
            final DebugProcess process = myDebuggerManager.getDebugProcess(p);
            if (process != null) {
                process.putUserData(XsltBreakpoint.MANAGER, null);
            }
        }
    }

    public static XsltBreakpointManager getInstance(Project project) {
        return project.getComponent(XsltBreakpointManager.class);
    }

    private class MyMouseAdapter extends EditorMouseAdapter {
        private boolean isFromMyProject(Editor editor) {
            FileEditor[] allEditors = FileEditorManager.getInstance(myProject).getAllEditors();
            for (FileEditor ed : allEditors) {
                if (!(ed instanceof TextEditor)) {
                    continue;
                }
                if (((TextEditor)ed).getEditor().equals(editor)) {
                    return true;
                }
            }
            return false;
        }

        public void mouseClicked(final EditorMouseEvent e) {
            final Editor editor = e.getEditor();
            if (MarkupEditorFilterFactory.createIsDiffFilter().avaliableIn(editor)) return;

            if (e.getArea() == EditorMouseEventArea.LINE_MARKERS_AREA) {
                PsiDocumentManager.getInstance(myProject).commitAndRunReadAction(new Runnable() {
                    public void run() {
                        if (!isFromMyProject(editor)) {
                            return;
                        }
                        final MouseEvent mouseEvent = e.getMouseEvent();
                        final int line = editor.xyToLogicalPosition(mouseEvent.getPoint()).line;
                        if (line < 0 || line >= editor.getDocument().getLineCount()) {
                            return;
                        }
                        if (mouseEvent.isPopupTrigger()) {
                            return;
                        }
                        if (mouseEvent.getButton() != 1) {
                            return;
                        }

                        DebuggerInvocationUtil.invokeLater(myProject, new Runnable() {
                            public void run() {
                                toggleBreakpoint(e.getEditor(), line);
                            }
                        });
                    }
                });
            }
        }
    }
}
