/*
 * Copyright 2002-2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger;

import org.intellij.plugins.xsltDebugger.ui.DebuggerTabComponent;

import com.intellij.compiler.impl.CompilerUtil;
import com.intellij.diagnostic.logging.AdditionalTabComponent;
import com.intellij.execution.CantRunException;
import com.intellij.execution.configurations.AdditionalTabComponentManager;
import com.intellij.execution.configurations.JavaParameters;
import com.intellij.execution.process.ProcessHandler;
import com.intellij.execution.process.ProcessListener;
import com.intellij.ide.plugins.IdeaPluginDescriptor;
import com.intellij.ide.plugins.PluginManager;
import com.intellij.openapi.application.PathManager;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.extensions.PluginId;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.projectRoots.Sdk;
import com.intellij.openapi.util.Key;
import com.intellij.openapi.util.UserDataHolder;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.JarFileSystem;
import com.intellij.util.net.NetUtils;

import org.intellij.lang.xpath.xslt.run.XsltRunConfiguration;
import org.intellij.lang.xpath.xslt.run.XsltRunnerExtension;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.jar.Manifest;
import java.util.jar.Attributes;

/**
 * Extension for XPathView that hooks into the execution of XSLT-scripts. Manages the creation of the XSLT-Debugger UI
 * and ensures the debugged process is started with the required JVM properties.
 */
public class XsltDebuggerExtension extends XsltRunnerExtension {
    private static final Logger LOG = Logger.getInstance(XsltDebuggerExtension.class.getName());

    private static final Key<Integer> PORT = Key.create("PORT");
    private static final Key<Manifest> MANIFEST = Key.create("MANIFEST");

    protected boolean supports(XsltRunConfiguration config, boolean debugger) {
        return debugger && config.getOutputType() == XsltRunConfiguration.OutputType.CONSOLE; // ?
    }

    public ProcessListener createProcessListener(Project project, UserDataHolder extensionData) {
        return new DebugProcessListener(project, extensionData.getUserData(PORT));
    }

    public boolean createTabs(Project project, AdditionalTabComponentManager manager, AdditionalTabComponent outputConsole, ProcessHandler process) {
        manager.addAdditionalTabComponent(DebuggerTabComponent.create(project, outputConsole, process), "XSLT-Debugger");
        return true;
    }

    public void patchParameters(JavaParameters parameters, XsltRunConfiguration configuration, UserDataHolder extensionData) throws CantRunException {
        final XsltRunConfiguration.OutputType outputType = configuration.getOutputType();

        final Sdk jdk = configuration.getEffectiveJDK();
        assert jdk != null;

        final String ver = jdk.getVersionString();
        if (!(CompilerUtil.isOfVersion(ver, "1.5") ||
                CompilerUtil.isOfVersion(ver, "5.0") ||
                CompilerUtil.isOfVersion(ver, "1.6") ||
                CompilerUtil.isOfVersion(ver, "1.7"))) {

            throw new CantRunException("The XSLT Debugger can only be used with JDK 1.5+");
        }

        // TODO: fix and remove
        if (outputType != XsltRunConfiguration.OutputType.CONSOLE) {
            throw new CantRunException("XSLT Debugger requires Output Type == CONSOLE");
        }

        try {
            final int port = NetUtils.findAvailableSocketPort();
            parameters.getVMParametersList().defineProperty("xslt.debugger.port", String.valueOf(port));
            extensionData.putUserData(PORT, port);
        } catch (IOException e) {
            LOG.info(e);
            throw new CantRunException("Unable to find a free network port");
        }

        final char c = File.separatorChar;

        final PluginId pluginId = PluginManager.getPluginByClassName(getClass().getName());
        assert pluginId != null;
        final IdeaPluginDescriptor descriptor = PluginManager.getPlugin(pluginId);
        assert descriptor != null;

        final File pluginPath = descriptor.getPath();

        final File rtClasspath = new File(pluginPath, "lib" + c + "xslt-debugger-engine.jar");
        if (rtClasspath.exists()) {
            parameters.getClassPath().addTail(rtClasspath.getAbsolutePath());
        } else {
            final File classes = new File(pluginPath, "classes");
            if (!classes.exists()) {
                LOG.warn("Debugger's Runtime classes not found in " + rtClasspath.getAbsolutePath());
                throw new CantRunException("Runtime classes not found");
            }

            parameters.getClassPath().addTail(classes.getAbsolutePath());
        }

        parameters.getClassPath().addTail(new File(PathManager.getLibPath() + c + "trove4j.jar").getAbsolutePath());

        final String type = parameters.getVMParametersList().getPropertyValue("xslt.transformer.type");
        if ("saxon".equalsIgnoreCase(type)) {
            addSaxon(parameters, pluginPath);
        } else if ("xalan".equalsIgnoreCase(type)) {
            addXalan(parameters, pluginPath);
        } else if (type != null) {
            throw new CantRunException("Unsupported Transformer type '" + type + "'");
        } else if (parameters.getClassPath().getPathsString().toLowerCase().contains("xalan")) {
            final List<VirtualFile> files = parameters.getClassPath().getVirtualFiles();
            for (VirtualFile file : files) {
                if (file.getName().matches(".*xalan.*\\.jar")) {
                    final VirtualFile root = JarFileSystem.getInstance().getJarRootForLocalFile(file);
                    final VirtualFile manifestFile = root != null ? root.findFileByRelativePath("META-INF/MANIFEST.MF") : null;
                    if (manifestFile != null) {
                        try {
                            Manifest manifest = manifestFile.getUserData(MANIFEST);
                            if (manifest == null) {
                                manifest = new Manifest(manifestFile.getInputStream());
                                manifestFile.putUserData(MANIFEST, manifest);
                            }
                            final Attributes attributes = manifest.getAttributes("org/apache/xalan/");
                            final String version = attributes.getValue("Implementation-Version");
                            if (version != null) {
                                final String[] parts = version.split("\\.");
                                if (parts.length >= 2) {
                                    if (Integer.parseInt(parts[0]) >= 2 && Integer.parseInt(parts[1]) >= 6) {
                                        parameters.getVMParametersList().defineProperty("xslt.transformer.type", "xalan");
                                        break;
                                    }
                                }
                                LOG.info("Unsupported Xalan version: " + version);
                            } else {
                                LOG.info("No Xalan version information in " + file.getPath());
                            }
                        } catch (IOException e) {
                            LOG.warn("Unable to read manifest from " + file.getName(), e);
                        }
                    } else {
                        LOG.info("No manifest file in " + file.getPath());
                    }
                    break;
                }
            }
        }
        if (!parameters.getVMParametersList().hasProperty("xslt.transformer.type")) {
            // add saxon for backward-compatibility
            parameters.getVMParametersList().defineProperty("xslt.transformer.type", "saxon");
            addSaxon(parameters, pluginPath);
        }

        parameters.getVMParametersList().defineProperty("xslt.main", "org.intellij.plugins.xsltDebugger.rt.XSLTDebuggerMain");
    }

    private static void addXalan(JavaParameters parameters, File pluginPath) {
        final char c = File.separatorChar;
        File xalan = new File(pluginPath, "lib" + c + "rt" + c + "xalan.jar");
        if (!xalan.exists()) {
            xalan = new File(pluginPath, "lib" + c + "xalan.jar");
        }
        parameters.getClassPath().addTail(xalan.getAbsolutePath());
        parameters.getClassPath().addTail(new File(xalan.getParentFile(), "serializer.jar").getAbsolutePath());
    }

    private static void addSaxon(JavaParameters parameters, File pluginPath) {
        final char c = File.separatorChar;
        File saxon = new File(pluginPath, "lib" + c + "rt" + c + "saxon.jar");
        if (!saxon.exists()) {
            saxon = new File(pluginPath, "lib" + c + "saxon.jar");
        }
        parameters.getClassPath().addTail(saxon.getAbsolutePath());
    }
}
