/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger;

import org.intellij.plugins.xsltDebugger.rt.engine.Breakpoint;
import org.intellij.plugins.xsltDebugger.rt.engine.BreakpointManager;
import org.intellij.plugins.xsltDebugger.rt.engine.Debugger;
import org.intellij.plugins.xsltDebugger.rt.engine.DebuggerStoppedException;
import org.intellij.plugins.xsltDebugger.ui.DebuggerTabComponent;

import com.intellij.debugger.DebuggerManagerEx;
import com.intellij.debugger.SourcePosition;
import com.intellij.debugger.engine.DebugProcess;
import com.intellij.debugger.engine.DebugProcessAdapter;
import com.intellij.debugger.engine.SuspendContext;
import com.intellij.execution.process.ProcessHandler;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.SelectionModel;
import com.intellij.openapi.editor.colors.EditorColorsManager;
import com.intellij.openapi.editor.colors.EditorColorsScheme;
import com.intellij.openapi.editor.ex.DocumentEx;
import com.intellij.openapi.editor.markup.HighlighterLayer;
import com.intellij.openapi.editor.markup.RangeHighlighter;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.fileEditor.OpenFileDescriptor;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Key;
import com.intellij.openapi.util.Pair;
import com.intellij.openapi.util.TextRange;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.util.EventDispatcher;
import com.intellij.xdebugger.ui.DebuggerColors;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.net.MalformedURLException;
import java.net.SocketException;
import java.net.URI;
import java.net.URISyntaxException;
import java.rmi.RemoteException;
import java.util.EventListener;

/**
 * This is the main place that interacts with the debugged XSLT processor. Waits until the processor
 * hits a breakpoint, resumes execution, etc.
 */
public class XsltDebuggerSession {
    private static final Key<XsltDebuggerSession> DEBUGGER_SESSION = Key.create("DEBUGGER_SESSION");

    private final Project myProject;
    private final ProcessHandler myProcess;
    private final Debugger myClient;
    private final EventDispatcher<Listener> myEventDispatcher = EventDispatcher.create(Listener.class);

    private Pair<SelectionModel, TextRange> myCurrentSelection;
    private Pair<Editor, RangeHighlighter> myExecutionPointHighlighter;
    private Breakpoint myTempBreakpoint;

    private volatile boolean myVmActive = true;
    private volatile Debugger.State myState;
    private boolean myClosed;

    private XsltDebuggerSession(Project project, ProcessHandler process, Debugger client) {
        myProject = project;
        myProcess = process;
        myClient = client;

        DebuggerManagerEx.getInstanceEx(project).getDebugProcess(process).addDebugProcessListener(new DebugProcessAdapter() {
            @Override
            public void paused(SuspendContext suspendContext) {
                myVmActive = false;
                myEventDispatcher.getMulticaster().vmPaused();
            }

            @Override
            public void resumed(SuspendContext suspendContext) {
                myVmActive = true;
                myEventDispatcher.getMulticaster().vmResumed();
            }

            @Override
            public void processDetached(DebugProcess process, boolean closedByUser) {
                myVmActive = false;
                process.removeDebugProcessListener(this);
            }
        });
    }

    public void start() {
        myClient.start();
        myState = Debugger.State.RUNNING;

        final BreakpointManager breakpointManager = myClient.getBreakpointManager();
        final Listener multicaster = myEventDispatcher.getMulticaster();
        try {
            if (!myClient.waitForDebuggee()) {
                multicaster.debuggerStopped();
                return;
            }
            myState = Debugger.State.SUSPENDED;
            do {
                if (myState == Debugger.State.SUSPENDED) {
                    if (myTempBreakpoint != null) {
                        breakpointManager.removeBreakpoint(myTempBreakpoint);
                        myTempBreakpoint = null;
                    }
                    highlightCurrentPosition();
                    multicaster.debuggerSuspended();
                } else if (myState == Debugger.State.RUNNING) {
                    removeExecutionPointHighlighter();
                    multicaster.debuggerResumed();
                } else if (myState == Debugger.State.STOPPED) {
                    break;
                }
            } while ((myState = myClient.waitForStateChange(myState)) != null);

            multicaster.debuggerStopped();
        } catch (DebuggerStoppedException e) {
            multicaster.debuggerStopped();
        } catch (RuntimeException e) {
            if (e.getCause() instanceof RemoteException) {
                if (e.getCause().getCause() instanceof SocketException) {
                    multicaster.debuggerStopped();
                    return;
                }
            }
            throw e;
        } finally {
            myState = Debugger.State.STOPPED;
            close();
        }
    }

    public DebuggerTabComponent getTabComponent() {
        return DebuggerTabComponent.getInstance(myProcess);
    }

    public void addListener(Listener listener) {
        myEventDispatcher.addListener(listener);
    }

    public void removeListener(Listener listener) {
        myEventDispatcher.removeListener(listener);
    }

    public Debugger getClient() {
        return myClient;
    }

    public Debugger.State getCurrentState() {
        return myState;
    }

    public void pause() {
        myClient.pause();
    }

    public void resume() {
        myClient.resume();
    }

    public void stop() {
        myClient.stop(false);
    }

    public void stepOver() {
        myClient.step();
    }

    public void stepInto() {
        myClient.stepInto();
    }

    public boolean canRunTo(PsiFile file, Document document, int offset) {
        return XsltBreakpoint.getActualLineNumber(new MySourcePosition(file, document, offset)) != -1;
    }

    public void runTo(final PsiFile file, final Document document, final int offset) {
        assert myTempBreakpoint == null;

        final int lineNumber = XsltBreakpoint.getActualLineNumber(new MySourcePosition(file, document, offset));
        final String uri = VfsUtil.virtualToIoFile(file.getVirtualFile()).toURI().toASCIIString();
        myTempBreakpoint = myClient.getBreakpointManager().setBreakpoint(uri, lineNumber);

        resume();
    }

    private void highlightCurrentPosition() {
        if (!SwingUtilities.isEventDispatchThread()) {
            ApplicationManager.getApplication().invokeLater(new Runnable() {
                public void run() {
                    highlightCurrentPosition();
                }
            });
            return;
        }

        removeSelection();
        removeExecutionPointHighlighter();

        final Debugger.StyleFrame frame = myClient.getCurrentFrame();
        final int lineNumber = frame.getLineNumber() - 1;

        @Nullable final Editor editor = openLocation(myProject, frame.getURI(), lineNumber);
        if (editor != null) {
            final EditorColorsScheme scheme = EditorColorsManager.getInstance().getGlobalScheme();
            myExecutionPointHighlighter = Pair.create(editor, editor.getMarkupModel().addLineHighlighter(
                    lineNumber,
                    HighlighterLayer.SELECTION - 1,
                    scheme.getAttributes(DebuggerColors.EXECUTIONPOINT_ATTRIBUTES))
            );
        }
    }

    public void showSelectedFrame(@NotNull final Debugger.Frame frame) {
        if (!SwingUtilities.isEventDispatchThread()) {
            ApplicationManager.getApplication().invokeLater(new Runnable() {
                public void run() {
                    showSelectedFrame(frame);
                }
            });
            return;
        }

        removeSelection();

        final int lineIndex = frame.getLineNumber() > 0 ? frame.getLineNumber() - 1 : 0;
        final Editor editor = openLocation(myProject, frame.getURI(), lineIndex);

        if (editor == null || frame.getNext() == null) {
            return;
        }
        final DocumentEx doc = (DocumentEx)editor.getDocument();
        final int startOffset = doc.getLineStartOffset(lineIndex);
        final int endOffset = doc.getLineEndOffset(lineIndex) + doc.getLineSeparatorLength(lineIndex);
        myCurrentSelection = Pair.create(editor.getSelectionModel(),
                new TextRange(startOffset, endOffset));

        myCurrentSelection.first.setSelection(startOffset, endOffset);
    }

    private void removeSelection() {
        if (!SwingUtilities.isEventDispatchThread()) {
            ApplicationManager.getApplication().invokeLater(new Runnable() {
                public void run() {
                    removeSelection();
                }
            });
            return;
        }

        if (myCurrentSelection == null) {
            return;
        }
        final SelectionModel selectionModel = myCurrentSelection.first;
        final TextRange range = myCurrentSelection.second;
        if (selectionModel.getSelectionStart() == range.getStartOffset() &&
                selectionModel.getSelectionEnd() == range.getEndOffset()) {
            selectionModel.removeSelection();
            myCurrentSelection = null;
        }
    }

    @Nullable
    public static Editor openLocation(Project project, @NotNull String uri, int lineNumber) {
        try {
            final VirtualFile file = VfsUtil.findFileByURL(new URI(uri).toURL());
            final OpenFileDescriptor descriptor = new OpenFileDescriptor(project, file, lineNumber, 0);
            descriptor.navigate(true);

            return FileEditorManager.getInstance(project).openTextEditor(descriptor, true);
        } catch (MalformedURLException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            return null;
        } catch (URISyntaxException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            return null;
        }
    }

    public synchronized void close() {
        if (myClosed) return;
        myClosed = true;

        removeExecutionPointHighlighter();
        removeSelection();

        XsltBreakpointManager.getInstance(myProject).detach(myProcess);
        XsltDebuggerSessionTracker.getInstance(myProject).notifySessionClosed(this);

        try {
            myClient.stop(true);
        } catch (DebuggerStoppedException e) {
            // OK
        } finally {
            myProcess.destroyProcess();
        }
    }

    private void removeExecutionPointHighlighter() {
        if (!SwingUtilities.isEventDispatchThread()) {
            ApplicationManager.getApplication().invokeLater(new Runnable() {
                public void run() {
                    removeExecutionPointHighlighter();
                }
            });
            return;
        }

        if (myExecutionPointHighlighter != null) {
            myExecutionPointHighlighter.first.getMarkupModel().removeHighlighter(myExecutionPointHighlighter.second);
            myExecutionPointHighlighter = null;
        }
    }

    @NotNull
    public static XsltDebuggerSession create(Project project, @NotNull ProcessHandler process, Debugger client) {
        final XsltDebuggerSession session = new XsltDebuggerSession(project, process, client);
        process.putUserData(DEBUGGER_SESSION, session);
        XsltDebuggerSessionTracker.getInstance(project).setActiveSession(session);
        return session;
    }

    public static XsltDebuggerSession getInstance(@NotNull ProcessHandler process) {
        return process.getUserData(DEBUGGER_SESSION);
    }

    public static void detach(ProcessHandler processHandler) {
        processHandler.putUserData(DEBUGGER_SESSION, null);
    }

    public void deselected() {
        if (myState == Debugger.State.SUSPENDED) {
            removeExecutionPointHighlighter();
        }
    }

    public void selected() {
        if (myState == Debugger.State.SUSPENDED) {
            highlightCurrentPosition();
        }
    }

    public boolean isVmActive() {
        return myVmActive;
    }

    public interface Listener extends EventListener {
        void debuggerSuspended();

        void debuggerResumed();

        void debuggerStopped();

        void vmPaused();

        void vmResumed();
    }

    private static final class MySourcePosition extends SourcePosition {
        private final PsiFile myFile;
        private final Document myDocument;
        private final int myOffset;

        public MySourcePosition(PsiFile file, Document document, int offset) {
            myFile = file;
            myDocument = document;
            myOffset = document.getLineStartOffset(document.getLineNumber(offset));
        }

        @Nullable
        public PsiElement getElementAt() {
            final int offset = getOffset();
            if (offset < 0) {
                return null;
            }
            final PsiFile psiFile = getFile();
            return psiFile.findElementAt(offset);
        }

        @NotNull
        public PsiFile getFile() {
            return myFile;
        }

        public int getLine() {
            return myDocument.getLineNumber(myOffset);
        }

        public int getOffset() {
            return myOffset;
        }

        @Nullable
        public Editor openEditor(boolean requestFocus) {
            final VirtualFile file = myFile.getVirtualFile();
            assert file != null;
            return FileEditorManager.getInstance(myFile.getProject()).openTextEditor(new OpenFileDescriptor(myFile.getProject(), file), requestFocus);
        }

        public void navigate(boolean requestFocus) {
        }

        public boolean canNavigate() {
            return false;
        }

        public boolean canNavigateToSource() {
            return false;
        }
    }
}
