/*
 * Copyright 2002-2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger;

import org.intellij.plugins.xsltDebugger.actions.XsltDebuggerActionManager;

import com.intellij.execution.ExecutionManager;
import com.intellij.execution.process.ProcessHandler;
import com.intellij.execution.executors.DefaultDebugExecutor;
import com.intellij.execution.ui.RunContentDescriptor;
import com.intellij.execution.ui.RunContentListener;
import com.intellij.execution.ui.RunContentManager;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.project.Project;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Keeps track of the currently focused XSLT-Debugger tab to be able route keyboard events for
 * Resume, Step, etc. to the correct process.
 */
public class XsltDebuggerSessionTracker implements ProjectComponent {
    private final Object myLock = new Object();

    private final Project myProject;
    private final XsltDebuggerActionManager myActionMgr;

    volatile XsltDebuggerSession myActiveSession;

    private final RunContentListener myContentListener = new RunContentListener() {
        public void contentSelected(RunContentDescriptor descriptor) {
            if (descriptor == null) {
                return;
            }

            synchronized (myLock) {
                final ProcessHandler process = descriptor.getProcessHandler();
                if (process != null) {
                    final XsltDebuggerSession session = XsltDebuggerSession.getInstance(process);
                    if (session != null) {
                        if (session != myActiveSession) {
                            if (myActiveSession != null) {
                                myActiveSession.deselected();
                            }

                            session.selected();
                            myActiveSession = session;
                        }
                    } else {
                        myActiveSession = null;
                    }
                } else {
                    myActiveSession = null;
                }
            }
        }

        public void contentRemoved(RunContentDescriptor descriptor) {
            if (descriptor == null) {
                return;
            }

            synchronized (myLock) {
                final ProcessHandler process = descriptor.getProcessHandler();
                if (process != null) {
                    final XsltDebuggerSession session = XsltDebuggerSession.getInstance(process);
                    if (session != null) {
                        if (session == myActiveSession) {
                            myActiveSession = null;
                        }
                        XsltDebuggerSession.detach(process);
                    }
                }
            }
        }
    };

    XsltDebuggerSessionTracker(Project project, XsltDebuggerActionManager actionMgr) {
        myProject = project;
        myActionMgr = actionMgr;
    }

    public static XsltDebuggerSessionTracker getInstance(Project project) {
        return project.getComponent(XsltDebuggerSessionTracker.class);
    }

    @Nullable
    public XsltDebuggerSession getActiveSession() {
        return myActiveSession;
    }

    public void initComponent() {
    }

    public void disposeComponent() {
    }

    @NotNull
    public String getComponentName() {
        return "XsltDebuggerSessionTracker";
    }

    public void projectOpened() {
        RunContentManager contentManager = ExecutionManager.getInstance(myProject).getContentManager();
        contentManager.addRunContentListener(myContentListener, DefaultDebugExecutor.getDebugExecutorInstance());
    }

    public void projectClosed() {
        final RunContentManager contentManager = ExecutionManager.getInstance(myProject).getContentManager();
        contentManager.removeRunContentListener(myContentListener);
        myActiveSession = null;
    }

    void setActiveSession(@NotNull XsltDebuggerSession session) {
        synchronized (myLock) {
            myActionMgr.sessionCreated();

            if (myActiveSession != null && myActiveSession != session) {
                myActiveSession.deselected();
            }
            myActiveSession = session;
        }
    }

    public synchronized void notifySessionClosed(XsltDebuggerSession session) {
        synchronized (myLock) {
            myActionMgr.sessionClosed();

            if (session == myActiveSession) {
                myActiveSession.deselected();
                myActiveSession = null;
            }
        }
    }
}
