/*
 * Copyright 2002-2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger.actions;

import org.intellij.plugins.xsltDebugger.XsltBreakpointManager;

import com.intellij.debugger.actions.DebuggerActions;
import com.intellij.openapi.actionSystem.ActionManager;
import com.intellij.openapi.actionSystem.AnAction;
import com.intellij.openapi.actionSystem.AnActionEvent;
import com.intellij.openapi.actionSystem.DataKeys;
import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.project.Project;
import com.intellij.util.NotNullFunction;
import org.jetbrains.annotations.NotNull;

import java.util.HashSet;
import java.util.Set;


/*
 * Created by IntelliJ IDEA.
 * User: sweinreuter
 * Date: 29.11.2007
 */
public class XsltDebuggerActionManager implements ApplicationComponent {
    private static final Logger LOG = Logger.getInstance(XsltDebuggerActionManager.class.getName());

    private static final String TOGGLE_LINE_BREAKPOINT = "ToggleLineBreakpoint";

    private static final NotNullFunction<AnAction,AnAction> UNHOOK_FUNC = new NotNullFunction<AnAction, AnAction>() {
        @NotNull
        public AnAction fun(AnAction anAction) {
            if (anAction instanceof HookedDebuggerAction) {
                return ((HookedDebuggerAction)anAction).getOrigAction();
            }
            return anAction;
        }
    };

    private final Set<String> myInstalledActions = new HashSet<String>();
    private int mySessionCount;

    public void initComponent() {
        final ActionManager manager = ActionManager.getInstance();

        install(manager, TOGGLE_LINE_BREAKPOINT, new NotNullFunction<AnAction, AnAction>() {
            @NotNull
            public AnAction fun(AnAction anAction) {
                return new MyToggleAction(anAction);
            }
        });
    }

    public void disposeComponent() {
        uninstall(ActionManager.getInstance(), TOGGLE_LINE_BREAKPOINT);
        uninstallWrappers();
    }

    public void sessionClosed() {
        synchronized (myInstalledActions) {
            mySessionCount--;
            LOG.assertTrue(mySessionCount >= 0);

            if (mySessionCount == 0) {
                uninstallWrappers();
                LOG.info("Debugger Control actions uninstalled");
            }
        }
    }

    public void sessionCreated() {
        synchronized (myInstalledActions) {
            if (++mySessionCount == 1) {
                LOG.assertTrue(myInstalledActions.size() == 0);

                final ActionManager manager = ActionManager.getInstance();
                installWrapper(manager, DebuggerActions.STEP_OVER, XsltDebuggerAction.STEP_OVER);
                installWrapper(manager, DebuggerActions.STEP_INTO, XsltDebuggerAction.STEP_INTO);
                installWrapper(manager, DebuggerActions.RUN_TO_CURSOR, XsltDebuggerAction.RUN_TO_CURSOR);
                installWrapper(manager, DebuggerActions.RESUME, XsltDebuggerAction.RESUME);

                LOG.info("Debugger Control actions installed");
            }
        }
    }

    private void install(ActionManager manager, String id, NotNullFunction<AnAction, AnAction> f) {
        final AnAction origAction = manager.getAction(id);
        manager.unregisterAction(id);
        manager.registerAction(id, f.fun(origAction));
    }

    private void uninstallWrappers() {
        final ActionManager manager = ActionManager.getInstance();
        for (String id : myInstalledActions) {
            uninstall(manager, id);
        }
        myInstalledActions.clear();
    }

    private void uninstall(ActionManager manager, String id) {
        install(manager, id, UNHOOK_FUNC);
    }

    private void installWrapper(ActionManager manager, String id, String targetId) {
        final AnAction targetAction = manager.getAction(targetId);
        install(manager, id, new NotNullFunction<AnAction, AnAction>() {
            @NotNull
            public AnAction fun(AnAction anAction) {
                return new DelegatingHookedAction(anAction, targetAction);
            }
        });
        myInstalledActions.add(id);
    }

    @NotNull
    public String getComponentName() {
        return "XSLT-Debugger.ActionInstaller";
    }

    private static class MyToggleAction extends HookedDebuggerAction {
        public MyToggleAction(AnAction origAction) {
            super(origAction);
        }

        protected boolean isEnabled(AnActionEvent e) {
            final Project project = e.getData(DataKeys.PROJECT);
            if (project == null) return false;
            final Editor editor = e.getData(DataKeys.EDITOR);
            return editor != null && XsltBreakpointManager.getInstance(project).canSetXsltBreakpoint(editor);
        }

        protected boolean actionPerformedImpl(AnActionEvent e) {
            final Project project = e.getData(DataKeys.PROJECT);
            if (project == null) return false;
            final Editor editor = e.getData(DataKeys.EDITOR);
            return  XsltBreakpointManager.getInstance(project).toggleBreakpoint(editor);
        }
    }

    private static class DelegatingHookedAction extends HookedDebuggerAction {
        private final AnAction myDelegate;

        public DelegatingHookedAction(AnAction action, AnAction delegate) {
            super(action);
            myDelegate = delegate;
        }

        protected boolean actionPerformedImpl(AnActionEvent e) {
            myDelegate.update(e);
            if (!e.getPresentation().isEnabled()) {
                return false;
            }
            myDelegate.actionPerformed(e);
            return true;
        }

        protected boolean isEnabled(AnActionEvent e) {
            if (!e.getPresentation().isEnabled()) {
                myDelegate.update(e);
                return e.getPresentation().isEnabled();
            }
            return true;
        }
    }
}
