/*
 * Copyright 2002-2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger.ui;

import org.intellij.plugins.xsltDebugger.rt.engine.Debugger;

import com.intellij.openapi.application.ApplicationManager;

import javax.swing.*;
import java.util.ArrayList;

/**
 * Created by IntelliJ IDEA.
* User: sweinreuter
* Date: 31.05.2007
*/
abstract class AbstractDataModel<F extends Debugger.Frame, T> extends AbstractListModel {
    private final java.util.List<T> myData = new ArrayList<T>();

    public synchronized int getSize() {
        return myData.size();
    }

    public synchronized T getElementAt(int index) {
        return myData.get(index);
    }

    public void update(F currentFrame) {
        final int oldSize;
        final int newSize;
        synchronized (this) {
            oldSize = getSize();
            
            myData.clear();

            fill(currentFrame);

            newSize = getSize();
        }

        if (!SwingUtilities.isEventDispatchThread()) {
            ApplicationManager.getApplication().invokeLater(new Runnable() {
                public void run() {
                    update(oldSize, newSize);
                }
            });
            return;
        }
        update(oldSize, newSize);
    }

    private void update(int oldSize, int newSize) {
        if (oldSize == newSize) {
            fireContentsChanged(this, 0, newSize);
        } else if (oldSize < newSize) {
            fireIntervalAdded(this, oldSize, newSize);
            fireContentsChanged(this, 0, oldSize);
        } else if (oldSize > newSize) {
            fireIntervalRemoved(this, oldSize, newSize);
            fireContentsChanged(this, 0, oldSize);
        }
    }

    protected void add(T t) {
        myData.add(t);
    }

    protected abstract void fill(F currentFrame);

    public void reset() {
        final int size;
        synchronized (this) {
            size = getSize();
            myData.clear();
        }

        if (!SwingUtilities.isEventDispatchThread()) {
            ApplicationManager.getApplication().invokeLater(new Runnable() {
                public void run() {
                    fireIntervalRemoved(AbstractDataModel.this, 0, size);
                }
            });
            return;
        }
        fireIntervalRemoved(this, 0, size);
    }
}
