/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger.ui;

import org.intellij.plugins.xsltDebugger.XsltDebuggerSession;
import org.intellij.plugins.xsltDebugger.actions.XsltDebuggerAction;
import org.intellij.plugins.xsltDebugger.rt.engine.Debugger;
import org.intellij.plugins.xsltDebugger.rt.engine.Debugger.Frame;
import org.intellij.plugins.xsltDebugger.rt.engine.OutputEventQueue;
import org.intellij.plugins.xsltDebugger.ui.actions.HideWhitespaceAction;
import org.intellij.plugins.xsltDebugger.ui.actions.OpenOutputAction;

import com.intellij.diagnostic.logging.AdditionalTabComponent;
import com.intellij.execution.process.ProcessHandler;
import com.intellij.ide.CommonActionsManager;
import com.intellij.openapi.actionSystem.ActionGroup;
import com.intellij.openapi.actionSystem.ActionManager;
import com.intellij.openapi.actionSystem.AnAction;
import com.intellij.openapi.actionSystem.AnActionEvent;
import com.intellij.openapi.actionSystem.CommonShortcuts;
import com.intellij.openapi.actionSystem.DefaultActionGroup;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Disposer;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.util.Key;
import com.intellij.ui.content.Content;
import com.intellij.ui.content.ContentFactory;
import com.intellij.util.Icons;
import com.intellij.util.ui.Tree;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import java.awt.*;

/**
 * Created by IntelliJ IDEA.
 * User: sweinreuter
 * Date: 30.05.2007
 */
public class DebuggerTabComponent extends AdditionalTabComponent implements XsltDebuggerSession.Listener {
    private static final Logger LOG = Logger.getInstance("org.intellij.plugins.xsltDebugger.ui.DebuggerTabComponent");

    private static final Key<DebuggerTabComponent> TAB_COMPONENT_KEY = Key.create("DebuggerTabComponent");

    public final Object myLock = new Object();

    private JComponent myComponent;

    private final JList myFramesList;
    private final MyFramesModel<Debugger.StyleFrame> myFramesModel;
    private final MyFramesModel<Debugger.SourceFrame> mySourceFramesModel;
    private final VariablesModel myVariablesModel;
    private final WatchesModel myWatchesModel;
    private final GeneratedStructureModel myEventModel;

    private XsltDebuggerSession mySession;

    private DebuggerTabComponent(final Project project, final AdditionalTabComponent console) {
        super(new BorderLayout());

        final ContentFactory contentFactory = ContentFactory.SERVICE.getInstance();

        final XsltDebuggerUI debuggerUI = new XsltDebuggerUI(project);
        Disposer.register(this, debuggerUI);
        Disposer.register(this, console);

        myFramesModel = new MyFramesModel<Debugger.StyleFrame>();
        myFramesList = new JList(myFramesModel);
        myFramesList.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(final ListSelectionEvent e) {
                if (!e.getValueIsAdjusting()) {
                    changeSelectedFrame();
                }
            }
        });
        myFramesList.setCellRenderer(new FrameCellRenderer());

        final Content frames = contentFactory.createContent(new JScrollPane(myFramesList), "Frames", false);
        frames.setIcon(IconLoader.getIcon("/debugger/frame.png"));

        final ActionManager mgr = ActionManager.getInstance();
        final ActionGroup executionActions = (ActionGroup)mgr.getAction(XsltDebuggerAction.CONTROL_GROUP);
        frames.setActions(executionActions, "XSLT.Debugger.Frames", null);
        debuggerUI.setFramesContent(frames);

        mySourceFramesModel = new MyFramesModel<Debugger.SourceFrame>();
        final JList sourceFramesList = new JList(mySourceFramesModel);
        sourceFramesList.setCellRenderer(new FrameCellRenderer());
        final Content sourceFrames = contentFactory.createContent(new JScrollPane(sourceFramesList), "Source", false);

        sourceFrames.setActions(executionActions, "XSLT.Debugger.SourceFrames", null);
        debuggerUI.setSourceFramesContent(sourceFrames);

        final AnAction showSelectedFrameAction = new AnAction() {
            public void update(AnActionEvent e) {
                e.getPresentation().setEnabled(sourceFramesList.getSelectedValue() != null);
            }

            public void actionPerformed(AnActionEvent e) {
                final Debugger.SourceFrame value = (Debugger.SourceFrame)sourceFramesList.getSelectedValue();
                mySession.showSelectedFrame(value);
            }
        };
        showSelectedFrameAction.registerCustomShortcutSet(CommonShortcuts.DOUBLE_CLICK_1, sourceFramesList);
        showSelectedFrameAction.registerCustomShortcutSet(CommonShortcuts.getEditSource(), sourceFramesList);

        myVariablesModel = new VariablesModel(project);
        final Tree variablesTree = new VariablesTree(myVariablesModel);

        final Content variables = contentFactory.createContent(new JScrollPane(variablesTree), "Variables", false);

        final DefaultActionGroup variableActions = new DefaultActionGroup();
        variableActions.add(mgr.getAction(XsltDebuggerAction.EVALUATE_EXPRESSION));
        variables.setIcon(IconLoader.getIcon("/debugger/value.png"));
        variables.setActions(variableActions, "XSLT.Debugger.Variables", null);
        debuggerUI.setVariablesContent(variables);

        final Content output = contentFactory.createContent(console.getComponent(), "Output", false);
        output.setIcon(IconLoader.getIcon("/debugger/console.png"));

        final DefaultActionGroup outputAtions = new DefaultActionGroup();
        outputAtions.add(new OpenOutputAction(console));
        output.setActions(outputAtions, "XSLT.Debugger.Output", null);

        debuggerUI.setOutputContent(output);

        myEventModel = new GeneratedStructureModel();
        final Tree structureTree = new StructureTree(myEventModel);
        myEventModel.addTreeModelListener(new SmartStructureTracker(structureTree, this));

        final Content structure = contentFactory.createContent(new JScrollPane(structureTree), "Structure", false);
        structure.setIcon(Icons.FLATTEN_PACKAGES_ICON);
        
        final DefaultActionGroup structureActions = new DefaultActionGroup();
        final StructureTreeExpander expander = new StructureTreeExpander(structureTree);
        final CommonActionsManager actionsManager = CommonActionsManager.getInstance();
        structureActions.add(new HideWhitespaceAction(structureTree, myEventModel));
        structureActions.add(actionsManager.createExpandAllAction(expander, structureTree));
        structureActions.add(actionsManager.createCollapseAllAction(expander, structureTree));
        structure.setActions(structureActions, "XSLT.Debugger.GeneratedStructure", null);

        debuggerUI.setOutputContent(structure);

        myWatchesModel = new WatchesModel(project);
        final VariablesTree watchesTree = new VariablesTree(myWatchesModel);
        final Content watches = contentFactory.createContent(new JScrollPane(watchesTree), "Watches", false);
        watches.setIcon(IconLoader.getIcon("/debugger/watches.png"));

        final ActionGroup watchesGroup = (ActionGroup)mgr.getAction(XsltDebuggerAction.WATCHES_GROUP);
        mgr.getAction(XsltDebuggerAction.EDIT_WATCH).registerCustomShortcutSet(CommonShortcuts.DOUBLE_CLICK_1, watchesTree);

        watches.setActions(watchesGroup, "XSLT-Debugger.Watches", null);
        debuggerUI.setVariablesContent(watches);

        myComponent = debuggerUI.getComponent();
        add(myComponent, BorderLayout.CENTER);
    }

    private void changeSelectedFrame() {
        synchronized (myLock) {
            final Debugger.StyleFrame frame = getSelectedFrame();
            if (frame != null) {
                mySession.showSelectedFrame(frame);
                myVariablesModel.update(frame);
            }
        }
    }


    @Nullable
    public Debugger.StyleFrame getSelectedFrame() {
        synchronized (myLock) {
            if (mySession == null) return null;
            if (mySession.getCurrentState() != Debugger.State.SUSPENDED) return null;

            final Debugger.StyleFrame frame = (Debugger.StyleFrame)myFramesList.getSelectedValue();
            return frame != null ? frame : mySession.getClient().getCurrentFrame();
        }
    }

    public String getTabTitle() {
        return "XSLT Debugger";
    }

    public JComponent getPreferredFocusableComponent() {
        return myComponent;
    }

    public void dispose() {
        synchronized (myLock) {
            reset();

            mySession = null;
            myComponent = null;
        }
    }

    void attachTo(XsltDebuggerSession session) {
        synchronized (myLock) {
            mySession = session;
            session.addListener(this);

// TODO fix this
//            Container grandparent = getParent().getParent();
//            if (grandparent instanceof ContentGroup) {
//                final JBTabs t = ((ContentGroup)grandparent).getTabs();
//                t.setSelected(t.findInfo(getTabTitle()), true);
//            }
        }
    }

    public static DebuggerTabComponent create(Project project, AdditionalTabComponent console, ProcessHandler startedProcess) {
        final DebuggerTabComponent tabComponent = new DebuggerTabComponent(project, console);
        startedProcess.putUserData(TAB_COMPONENT_KEY, tabComponent);
        return tabComponent;
    }

    public static DebuggerTabComponent getInstance(ProcessHandler process) {
        return process.getUserData(TAB_COMPONENT_KEY);
    }

    public static void attachDebugger(ProcessHandler handler, XsltDebuggerSession session) {
        handler.getUserData(TAB_COMPONENT_KEY).attachTo(session);
    }

    public void debuggerSuspended() {
        synchronized (myLock) {
            update();
        }
    }

    public void debuggerResumed() {
        synchronized (myLock) {
            reset();
        }
    }

    public void debuggerStopped() {
        synchronized (myLock) {
            reset();

            final Debugger client = mySession.getClient();

            mySession.removeListener(this);
            mySession = null;

            try {
                final OutputEventQueue queue = client.getEventQueue();
                myEventModel.finalUpdate(queue.getEvents());
            } catch (Exception e) {
                // can fail when debugger is manually terminated
            }
        }
    }

    public void vmPaused() {
        synchronized (myLock) {
            reset();
        }
    }

    public void vmResumed() {
        synchronized (myLock) {
            if (mySession.getCurrentState() == Debugger.State.SUSPENDED) {
                update();
            }
        }
    }

    private void update() {
        assert Thread.holdsLock(myLock);

        final Debugger client = mySession.getClient();
        final Debugger.StyleFrame frame = client.getCurrentFrame();

        myFramesModel.update(frame);
        mySourceFramesModel.update(client.getSourceFrame());

        myVariablesModel.update(frame);
        myWatchesModel.update(frame);
        
        myEventModel.update(client.getEventQueue().getEvents());
    }

    private void reset() {
        assert Thread.holdsLock(myLock);

        myFramesModel.reset();
        mySourceFramesModel.reset();

        myVariablesModel.reset();
        myWatchesModel.reset();
    }

    public WatchesModel getWatchesModel() {
        return myWatchesModel;
    }

    public VariablesModel getVariablesModel() {
        return myVariablesModel;
    }

    public ActionGroup getToolbarActions() {
        // TODO
        return null;
    }

    public JComponent getSearchComponent() {
        // TODO
        return null;
    }

    public String getToolbarPlace() {
        // TODO
        return null;
    }

    public JComponent getToolbarContextComponent() {
        // TODO
        return null;
    }

    public boolean isContentBuiltIn() {
        // TODO
        return false;
    }

    private static class MyFramesModel<F extends Frame<F>> extends AbstractDataModel<F, Debugger.Frame> {
        protected void fill(F currentFrame) {
            while (currentFrame != null) {
                add(currentFrame);
                currentFrame = currentFrame.getPrevious();
            }
        }
    }
}
