/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger.ui;

import org.intellij.plugins.xsltDebugger.rt.engine.Debugger;
import org.intellij.plugins.xsltDebugger.rt.engine.Value;

import com.intellij.ide.util.PropertiesComponent;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.event.DocumentAdapter;
import com.intellij.openapi.editor.event.DocumentEvent;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.ComboBox;
import com.intellij.openapi.ui.DialogWrapper;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiFileFactory;
import com.intellij.ui.EditorTextField;
import com.intellij.util.ArrayUtil;
import com.intellij.util.LocalTimeCounter;
import com.intellij.util.ui.Tree;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.Nullable;

import org.intellij.lang.xpath.XPathFileType;
import org.intellij.lang.xpath.context.ContextProvider;

import javax.swing.*;
import javax.swing.plaf.basic.BasicComboBoxEditor;
import java.awt.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

/**
 * Created by IntelliJ IDEA.
 * User: sweinreuter
 * Date: 06.06.2007
 */
public class EvaluationDialog extends DialogWrapper {
    private static final Logger LOG = Logger.getInstance("org.intellij.plugins.xsltDebugger.ui.DebuggerTabComponent");

    private JLabel myIcon;
    private ComboBox myExpression;
    private JPanel myRoot;
    private Tree myResultTree;
    private JPanel myResultPanel;

    private final EditorTextField myEditor;
    private String[] myHistory;
    private final Debugger.StyleFrame myFrame;
    private final VariablesModel myResultModel;
    private final Project myProject;

    public EvaluationDialog(Project project, String selection, ContextProvider provider, String title) {
        this(project, selection, provider, (Debugger.StyleFrame)null);
        setTitle(title);
    }

    public EvaluationDialog(Project project, String selection, ContextProvider provider, Debugger.StyleFrame frame) {
        super(project, false);
        myProject = project;
        myFrame = frame;

        // inline evaluation only available if frame is present
        myResultPanel.setVisible(frame != null);

        final PsiFileFactory factory = PsiFileFactory.getInstance(project);

        final PsiFile file = factory.createFileFromText("XPathExpr.xpath", XPathFileType.XPATH, "", LocalTimeCounter.currentTime(), true);
        provider.attachTo(file);
        final Document document = PsiDocumentManager.getInstance(project).getDocument(file);

        myEditor = new EditorTextField(document, project, XPathFileType.XPATH);
        myHistory = loadHistory();
        myExpression.setModel(new DefaultComboBoxModel(myHistory));

        myExpression.setEditor(new EditorAdapter(myEditor));
        myExpression.setEditable(true);
        myEditor.setText(selection != null ? selection : "");
        myEditor.addDocumentListener(new DocumentAdapter() {
            @Override
            public void documentChanged(DocumentEvent e) {
                setOKActionEnabled(e.getDocument().getTextLength() > 0);
            }
        });
        setOKActionEnabled(selection != null && selection.length() > 0);
        
        myIcon.setIcon(Messages.getQuestionIcon());
        myIcon.setText(null);

        myResultModel = new VariablesModel(project);
        myResultTree.setModel(myResultModel);

        setTitle("Evaluate Expression");

        init();
    }

    private void createUIComponents() {
        myResultTree = new VariablesTree();
    }

    @Override
    public JComponent getPreferredFocusedComponent() {
        return myEditor.getComponent();
    }

    protected String[] loadHistory() {
        final PropertiesComponent propertiesComponent = PropertiesComponent.getInstance();
        final String history = propertiesComponent.getValue(getPropertyKey());
        if (history != null) {
            final String[] items = history.split("\n");
            ArrayList<String> result = new ArrayList<String>();
            for (String item : items) {
                if (item != null && item.length() > 0) {
                    result.add(item);
                }
            }
            return result.toArray(new String[result.size()]);
        }
        return ArrayUtil.EMPTY_STRING_ARRAY;
    }

    @Override
    protected void doOKAction() {
        myHistory = saveHistory();
        myExpression.setModel(new DefaultComboBoxModel(myHistory));

        myResultModel.reset();

        if (myFrame != null) {
            final String s = getExpression();
            try {
                final Value value = myFrame.eval(s);
                myResultModel.update(Collections.<Debugger.Variable>singletonList(new MyResult(value)));
            } catch (Debugger.EvaluationException e1) {
                LOG.debug(e1);
                Messages.showErrorDialog(myRoot, "Failed to evaluate expression: " + e1.getMessage(), "Evaluation Failed");
            }
        } else {
            super.doOKAction();
        }
    }

    @Override
    protected Action getOKAction() {
        if (myFrame != null) {
            final Action action = super.getOKAction();
            action.putValue(Action.NAME, "Evaluate");
            return action;
        }
        return super.getOKAction();
    }

    @Override
    protected Action getCancelAction() {
        if (myFrame != null) {
            final Action action = super.getCancelAction();
            action.putValue(Action.NAME, "Close");
            return action;
        }
        return super.getCancelAction();
    }

    protected String[] saveHistory() {
        final PropertiesComponent propertiesComponent = PropertiesComponent.getInstance(myProject);
        final String propertyKey = getPropertyKey();
        final String text = getExpression();

        if (text.trim().length() > 0) {
            final ArrayList<String> history = new ArrayList<String>(Arrays.asList(myHistory));
            history.remove(text);
            history.add(0, text);
            propertiesComponent.setValue(propertyKey, StringUtil.join(history, "\n"));
            return history.toArray(new String[history.size()]);
        }
        return myHistory;
    }

    protected String getPropertyKey() {
        return getDimensionKey() + ".HISTORY";
    }

    @Override
    @NonNls
    protected String getDimensionServiceKey() {
        return myFrame != null ? "XSLT.Debugger.ExpressionEvaluation" : "XSLT.Debugger.EditWatch";
    }

    @Nullable
    protected JComponent createCenterPanel() {
        return myRoot;
    }

    public String getExpression() {
        return myEditor.getText();
    }

    protected class EditorAdapter extends BasicComboBoxEditor {
        private final EditorTextField myTf;

        public EditorAdapter(EditorTextField tf) {
            myTf = tf;
        }

        public Component getEditorComponent() {
            return myTf.getComponent();
        }

        @Nullable
        public Object getItem() {
            return myExpression.getModel().getSelectedItem();
        }

        public void selectAll() {
            myTf.selectAll();
        }

        public void setItem(Object object) {
            if (object == null) {
                myEditor.setText("");
            } else {
                myEditor.setText((String)object);
            }
        }
    }

    private static class MyResult implements Debugger.Variable {
        private final Value myValue;

        public MyResult(Value value) {
            myValue = value;
        }

        @SuppressWarnings({ "ConstantConditions" })
        public String getURI() {
            return null;
        }

        public int getLineNumber() {
            return -1;
        }

        public boolean isGlobal() {
            return false;
        }

        public Kind getKind() {
            return Kind.EXPRESSION;
        }

        public String getName() {
            return "result";
        }

        public Value getValue() {
            return myValue;
        }
    }
}
