/*
 * Copyright 2002-2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger.ui;

import org.intellij.plugins.xsltDebugger.XsltDebuggerSession;
import org.intellij.plugins.xsltDebugger.rt.engine.Debugger;
import org.intellij.plugins.xsltDebugger.rt.engine.Value;

import com.intellij.debugger.settings.NodeRendererSettings;
import com.intellij.debugger.ui.tree.render.ArrayRenderer;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.pom.Navigatable;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.*;
import java.util.Collections;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
* User: sweinreuter
* Date: 07.06.2007
*/
public class VariablesModel extends DefaultTreeModel {

    public VariablesModel(Project project) {
        super(new MyRootNode(project));
    }

    @Override
    public MyRootNode getRoot() {
        return (MyRootNode)super.getRoot();
    }

    @SuppressWarnings({ "unchecked" })
    public synchronized void update(Debugger.StyleFrame frame) {
        final MyRootNode root = getRoot();
        root.setUserObject(frame);
        root.removeAllChildren();

        final List<Debugger.Variable> variables = frame.getVariables();
        update(variables);
    }

    public void update(List<Debugger.Variable> variables) {
        final MyRootNode root = getRoot();

        for (Debugger.Variable variable : variables) {
            final Value value = variable.getValue();

            if (value.getType() == Value.XPathType.NODESET) {
                final Object v = value.getValue();
                if (v instanceof Value.NodeSet) {
                    final Value.NodeSet nodeset = (Value.NodeSet)v;
                    final NodeSetNode childNode = new NodeSetNode(variable, nodeset);
                    root.add(childNode);

                    final List<Value.Node> nodes = nodeset.getNodes();
                    final ArrayRenderer renderer = NodeRendererSettings.getInstance().getArrayRenderer();
                    int limit = renderer.ENTRIES_LIMIT;
                    if (nodes.size() > limit) {
                        limit >>= 1;
                        for (int i = 0; i < limit; i++) {
                            final Value.Node node = nodes.get(i);
                            childNode.add(new NodeInfoNode(node));
                        }
                        final int max = nodes.size() - 1;
                        for (int i = max - limit; i < max; i++) {
                            final Value.Node node = nodes.get(i);
                            childNode.add(new NodeInfoNode(node));
                        }
                    } else {
                        for (Value.Node node : nodes) {
                            childNode.add(new NodeInfoNode(node));
                        }
                    }
                    continue;
                }
            }

            final VariableNode child = new VariableNode(variable);
            root.add(child);
        }

        if (!SwingUtilities.isEventDispatchThread()) {
            ApplicationManager.getApplication().invokeLater(new Runnable() {
                public void run() {
                    nodeStructureChanged(root);
                }
            });
            return;
        }
        nodeStructureChanged(root);
    }

    public void reset() {
        final MyRootNode root = getRoot();
        root.removeAllChildren();
        
        if (!SwingUtilities.isEventDispatchThread()) {
            ApplicationManager.getApplication().invokeLater(new Runnable() {
                public void run() {
                    nodeStructureChanged(root);
                }
            });
            return;
        }
        nodeStructureChanged(root);
    }

    public synchronized List<Debugger.Variable> getVariables() {
        final Debugger.StyleFrame frame = getRoot().getFrame();
        return frame != null ? frame.getVariables() : Collections.<Debugger.Variable>emptyList();
    }

    private static final class MyRootNode extends DefaultMutableTreeNode {
        private final Project myProject;

        MyRootNode(Project project) {
            myProject = project;
        }

        public Debugger.StyleFrame getFrame() {
            return (Debugger.StyleFrame)getUserObject();
        }
    }

    public static class VariableNode extends DefaultMutableTreeNode implements Navigatable {
        VariableNode(Debugger.Variable variable) {
            super(variable);
        }

        @Override
        public Debugger.Variable getUserObject() {
            return (Debugger.Variable)super.getUserObject();
        }

        public void navigate(boolean requestFocus) {
            final Debugger.Variable variable = getUserObject();
            XsltDebuggerSession.openLocation(((MyRootNode)getPath()[0]).myProject, variable.getURI(), variable.getLineNumber() - 1);
        }

        public boolean canNavigate() {
            final Debugger.Variable v = getUserObject();
            return v.getLineNumber() > 0 && v.getURI() != null;
        }

        public boolean canNavigateToSource() {
            return canNavigate();
        }
    }

    public static class NodeSetNode extends VariableNode implements Navigatable {
        private final Value.NodeSet myNodeSet;

        NodeSetNode(Debugger.Variable variable, Value.NodeSet nodeset) {
            super(variable);
            myNodeSet = nodeset;
        }

        public String getStringValue() {
            return myNodeSet.myStringValue;
        }
    }

    public static class NodeInfoNode extends DefaultMutableTreeNode implements Navigatable {
        NodeInfoNode(Value.Node node) {
            super(node);
        }

        public Value.Node getNodeInfo() {
            return (Value.Node)getUserObject();
        }

        public void navigate(boolean requestFocus) {
            final Value.Node node = getNodeInfo();
            XsltDebuggerSession.openLocation(((MyRootNode)getPath()[0]).myProject, node.myURI, node.myLineNumber - 1);
        }

        public boolean canNavigate() {
            return getNodeInfo().myLineNumber > 0;
        }

        public boolean canNavigateToSource() {
            return canNavigate();
        }
    }
}
