/*
 * Copyright 2002-2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger.ui;

import org.intellij.plugins.xsltDebugger.rt.engine.Debugger;
import org.intellij.plugins.xsltDebugger.rt.engine.Value;

import com.intellij.openapi.project.Project;

import java.util.ArrayList;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: sweinreuter
 * Date: 19.06.2007
 */
public class WatchesModel extends VariablesModel {
    private final List<Debugger.Variable> myWatches = new ArrayList<Debugger.Variable>();

    public WatchesModel(Project project) {
        super(project);
    }

    public synchronized void addWatch(String expr, Debugger.StyleFrame selectedFrame) {
        final WatchExpression watch = new WatchExpression(expr);
        if (selectedFrame != null) {
            watch.compute(selectedFrame);
        }

        myWatches.add(watch);
        reset();
        update(myWatches);
    }

    @Override
    @SuppressWarnings({ "unchecked" })
    public synchronized void update(Debugger.StyleFrame frame) {
        for (Debugger.Variable watch : myWatches) {
            ((WatchExpression)watch).compute(frame);
        }
        update(myWatches);
    }

    public synchronized void removeWatch(Debugger.Variable watch) {
        myWatches.remove(watch);
        reset();
        update(myWatches);
    }

    public void editWatch(Debugger.Variable variable, String expression, Debugger.StyleFrame frame) {
        final WatchExpression watch = (WatchExpression)variable;
        watch.setExpression(expression);
        if (frame != null) {
            watch.compute(frame);

            reset();
            update(myWatches);
        }
    }

    private static class WatchExpression implements Debugger.Variable {
        private String myExpr;
        private Value myValue;

        public WatchExpression(String expr) {
            myExpr = expr;
        }

        public boolean isGlobal() {
            return false;
        }

        public Kind getKind() {
            return Kind.EXPRESSION;
        }

        public String getName() {
            return myExpr;
        }

        public Value getValue() {
            return myValue;
        }

        @SuppressWarnings({ "ConstantConditions" })
        public String getURI() {
            return null;
        }

        public int getLineNumber() {
            return -1;
        }

        public void compute(Debugger.StyleFrame frame) {
            try {
                myValue = frame.eval(myExpr);
            } catch (final Debugger.EvaluationException e) {
                myValue = new BadValue(e);
            }
        }

        public void setExpression(String expression) {
            myExpr = expression;
        }
    }

    public static class BadValue implements Value {
        private final Debugger.EvaluationException myException;

        public BadValue(Debugger.EvaluationException e) {
            myException = e;
        }

        public Object getValue() {
            return myException.getMessage();
        }

        public Type getType() {
            return new ObjectType(myException.getClass().getName());
        }
    }
}
