/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger.ui;

import org.intellij.plugins.xsltDebugger.XsltBreakpoint;

import com.intellij.codeInsight.intention.IntentionAction;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.psi.PsiFileFactory;
import com.intellij.psi.xml.XmlElement;
import com.intellij.ui.EditorTextField;
import com.intellij.util.LocalTimeCounter;
import com.intellij.xdebugger.impl.ui.DebuggerUIUtil;
import org.jetbrains.annotations.NotNull;

import org.intellij.lang.xpath.XPathFile;
import org.intellij.lang.xpath.XPathFileType;
import org.intellij.lang.xpath.context.VariableContext;
import org.intellij.lang.xpath.context.XPathQuickFixFactoryImpl;
import org.intellij.lang.xpath.psi.XPathVariableReference;
import org.intellij.lang.xpath.validation.inspections.quickfix.XPathQuickFixFactory;
import org.intellij.lang.xpath.xslt.context.XsltContextProvider;
import org.intellij.lang.xpath.xslt.context.XsltVariableContext;

import javax.swing.*;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

public class XsltBreakpointSettings {
    private final Project myProject;

    private JPanel myRoot;

    private JCheckBox mySuspend;

    private JCheckBox myExpression;
    private EditorTextField myConditionExpr;

    private JCheckBox myLogMessage;
    private EditorTextField myLogMessageExpr;

    private JCheckBox myTraceMessage;
    private EditorTextField myTraceMessageExpr;

    private PsiFile myFile;
    private PsiFile myFile2;
    private PsiFile myFile3;

    public XsltBreakpointSettings(Project project) {
        myProject = project;
        myExpression.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent e) {
                myConditionExpr.setEnabled(e.getStateChange() == ItemEvent.SELECTED);
            }
        });
        myLogMessage.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent e) {
                final boolean b = e.getStateChange() == ItemEvent.SELECTED;
                myLogMessageExpr.setEnabled(b);
                mySuspend.setEnabled(b || myTraceMessage.isSelected());
            }
        });
        myTraceMessage.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent e) {
                final boolean b = e.getStateChange() == ItemEvent.SELECTED;
                myTraceMessageExpr.setEnabled(b);
                mySuspend.setEnabled(b || myLogMessage.isSelected());
            }
        });
        DebuggerUIUtil.focusEditorOnCheck(myExpression, myConditionExpr);
        DebuggerUIUtil.focusEditorOnCheck(myLogMessage, myLogMessageExpr);
        DebuggerUIUtil.focusEditorOnCheck(myTraceMessage, myTraceMessageExpr);        
    }

    private void createUIComponents() {
        final PsiFileFactory factory = PsiFileFactory.getInstance(myProject);

        myFile = factory.createFileFromText("ConditionExpr.xpath", XPathFileType.XPATH, "", LocalTimeCounter.currentTime(), true);
        final Document document = PsiDocumentManager.getInstance(myProject).getDocument(myFile);
        myConditionExpr = new EditorTextField(document, myProject, XPathFileType.XPATH);

        myFile2 = factory.createFileFromText("LogMessageExpr.xpath", XPathFileType.XPATH, "", LocalTimeCounter.currentTime(), true);
        final Document document2 = PsiDocumentManager.getInstance(myProject).getDocument(myFile2);
        myLogMessageExpr = new EditorTextField(document2, myProject, XPathFileType.XPATH);

        myFile3 = factory.createFileFromText("TraceMessageExpr.xpath", XPathFileType.XPATH, "", LocalTimeCounter.currentTime(), true);
        final Document document3 = PsiDocumentManager.getInstance(myProject).getDocument(myFile3);
        myTraceMessageExpr = new EditorTextField(document3, myProject, XPathFileType.XPATH);
    }

    public JComponent getRoot() {
        return myRoot;
    }

    public void initFrom(XsltBreakpoint breakpoint) {
        final boolean b = breakpoint.HAS_CONDITION;
        myExpression.setSelected(b);
        myConditionExpr.setEnabled(b);
        myConditionExpr.setText(breakpoint.CONDITION_EXPR);

        final boolean b2 = breakpoint.HAS_LOG_MSG;
        myLogMessage.setSelected(b2);
        myLogMessageExpr.setEnabled(b2);
        myLogMessageExpr.setText(breakpoint.LOG_EXPR);

        final boolean b3 = breakpoint.HAS_TRACE_MSG;
        myTraceMessage.setSelected(b3);
        myTraceMessageExpr.setEnabled(b3);
        myTraceMessageExpr.setText(breakpoint.TRACE_EXPR);

        mySuspend.setEnabled(b2 || b3);
        mySuspend.setSelected((b2 || b3) && breakpoint.SUSPEND);

        final PsiElement contextElement = breakpoint.findContextElement();
        if (contextElement != null) {
            final BreakpointContext context = new BreakpointContext(contextElement);
            context.attachTo(myFile);
            context.attachTo(myFile2);
            context.attachTo(myFile3);
        }
    }

    public void saveTo(XsltBreakpoint breakpoint) {
        breakpoint.HAS_CONDITION = myExpression.isSelected();
        breakpoint.CONDITION_EXPR = myConditionExpr.getText();
        breakpoint.HAS_LOG_MSG = myLogMessage.isSelected();
        breakpoint.LOG_EXPR = myLogMessageExpr.getText();
        breakpoint.HAS_TRACE_MSG = myTraceMessage.isSelected();
        breakpoint.TRACE_EXPR = myTraceMessageExpr.getText();
        breakpoint.SUSPEND = mySuspend.isSelected();
    }

    private static class BreakpointContext extends XsltContextProvider {
        public BreakpointContext(PsiElement contextElement) {
            super((XmlElement)contextElement);
        }

        @Override
        @NotNull
        public XPathQuickFixFactory getQuickFixFactory() {
            return XPathQuickFixFactoryImpl.INSTANCE;
        }

        @Override
        public PsiFile[] getRelatedFiles(XPathFile file) {
            return PsiFile.EMPTY_ARRAY;
        }

        @Override
        @NotNull
        public VariableContext getVariableContext() {
            return new XsltVariableContext() {
//                @Nullable
//                protected XmlTag getContextTag(XPathElement element) {
//                    return PsiTreeUtil.getParentOfType(getContextElement(), XmlTag.class, false);
//                }

                @Override
                @NotNull
                public IntentionAction[] getUnresolvedVariableFixes(XPathVariableReference reference) {
                    return new IntentionAction[0];
                }
            };
        }
    }
}