/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.xsltDebugger.ui;

import com.intellij.ide.ui.SplitterProportionsDataImpl;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.wm.IdeFocusManager;
import com.intellij.openapi.actionSystem.ActionManager;
import com.intellij.openapi.project.Project;
import com.intellij.ui.GuiUtils;
import com.intellij.ui.content.Content;
import com.intellij.ui.tabs.JBTabs;
import com.intellij.ui.tabs.TabInfo;
import com.intellij.ui.tabs.impl.JBTabsImpl;
import org.jetbrains.annotations.NonNls;

import javax.swing.*;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * Created by IntelliJ IDEA.
 * User: sweinreuter
 * Date: 31.05.2007
 */
public class XsltDebuggerUI implements Disposable {
    @NonNls
    private static final String KEY = "XSLT.Debugger.UI.Proportions";

    private final SplitterProportionsDataImpl myProportionsData;
    private final Project myProject;

    private JPanel myRootPanel;
    private JBTabsImpl myFramesTab;
    private JBTabsImpl myVariablesTab;
    private JBTabsImpl myOutputTab;

    public XsltDebuggerUI(Project project) {
        myProject = project;
        myProportionsData = new SplitterProportionsDataImpl();
        myProportionsData.setProportions(new ArrayList<Float>(Arrays.asList(0.5f, 0.5f))); // default

        GuiUtils.replaceJSplitPaneWithIDEASplitter(myRootPanel);
    }

    private void createUIComponents() {
        myRootPanel = new JPanel() {
            @Override
            public void addNotify() {
                super.addNotify();
                Runnable runnable = new Runnable() {
                    public void run() {
                        synchronized (getClass()) {
                            myProportionsData.externalizeFromDimensionService(KEY);
                            myProportionsData.restoreSplitterProportions(myRootPanel);
                        }
                    }
                };
                SwingUtilities.invokeLater(runnable);
            }
        };

        final ActionManager actionManager = ActionManager.getInstance();
        // TODO use proper API here
        myFramesTab = new JBTabsImpl(myProject, actionManager, IdeFocusManager.getInstance(myProject), this);
        myVariablesTab = new JBTabsImpl(myProject, actionManager, IdeFocusManager.getInstance(myProject), this);
        myOutputTab = new JBTabsImpl(myProject, actionManager, IdeFocusManager.getInstance(myProject), this);
    }

    public void setFramesContent(Content frames) {
        createTab(myFramesTab, frames);
    }

    public void setSourceFramesContent(Content frames) {
        createTab(myFramesTab, frames);
    }

    public void setVariablesContent(Content frames) {
        createTab(myVariablesTab, frames);
    }

    public void setOutputContent(Content frames) {
        createTab(myOutputTab, frames);
    }

    private void createTab(JBTabs tab, Content frames) {
        final TabInfo info = new TabInfo(frames.getComponent());
        tab.addTab(info, tab.getTabCount());
        info.setText(frames.getDisplayName());
        info.setIcon(frames.getIcon());
        info.setActions(frames.getActions(), frames.getPlace());
    }

    public JComponent getComponent() {
        return myRootPanel;
    }

    public void dispose() {
        synchronized (getClass()) {
            myProportionsData.saveSplitterProportions(myRootPanel);
            myProportionsData.externalizeToDimensionService(KEY);
        }
    }
}
